// hovercraft.cpp
#include "hovercraft.h"
#include "data.h"
#include "manager.h"
#include "language.h"

#include <algorithm>
#include <iostream>
#include <sstream>

// Controller methods

void KeyboardController::tick(Component quantum)
{
	// TODO
	rudder = 0;
	if (key[KEY_LEFT]) rudder = -1;
	if (key[KEY_RIGHT]) rudder = 1;
	if (key[KEY_UP]) 
		power = 1; 
	else 
		power = 0;
}

Component KeyboardController::getRudder()
{
	return rudder;
}

Component KeyboardController::getPower()
{
	return power;
};

/// HOVERCRAFT METhods
static const Component Radius = 16.0;
// Bumper distance from centre
static const Component BumperDist = 6.0;
static const Component BumperRadius = 10.0;
// BumperRadius + BumperDist = Radius this is what we want.

HoverCraft::HoverCraft(Manager *m, int p) : GameObject(m),
	azimuth(0), velocity(Vect3d(0,0,0)), rotation(0), controller(0),
	playerNum(p)
{
	radius = Radius;
	// TODO: fix up our controller
	sprites = 0;
	velocityDelta = Vect3d(0,0,0);
	wakeSprite = 0;
	initSprites();
	currentLap = 0;
	nextCheckpoint = 0;
	currentLapBegin = 0;
	lastLapTime = 0;
	lastCheckpoint = -1;
	lastCheckpointTime = 0;
	courseComplete = false;
	isPlayer = false;
	clearPowerups();
	wakeTimer = 0;
}

void HoverCraft::modifyColour(int &r, int &g, int &b)
{
	switch (playerNum) {
		case 2: // Swap red/blue (Blue craft)
			std::swap(r,b);
			break;
		case 3: // Swap red/green (Green craft)
			std::swap(r,g);
			break;
		case 4: // Copy red to green (Yellow)
			g = r;
			break;
	}
}

void HoverCraft::initSprites()
{
	clearSprites();
	set_write_alpha_blender();
	sprites = new BITMAP *[36];
	for (int i=0; i<36; i++) {
		BITMAP *model = manager->getSprite(HOVER01_TGA + i);
		sprites[i] = create_bitmap_ex(32,model->w, model->h);
		blit(model, sprites[i], 0, 0, 0, 0, model->w, model->h);
		for (int x=0; x<model->w; x++) {
			for (int y=0; y<model->h; y++) {
				int col = getpixel(sprites[i],x,y);
				int r = getr32(col);
				int g = getg32(col);
				int b = getb32(col);
				modifyColour(r,g,b);
				int alpha = 0;
				if ((r+g+b) < 64) {
					alpha = (64 - r+g+b) *4;
					if (alpha > 255) alpha = 255;
				}
				alpha = 255-alpha;
				putpixel(sprites[i],x,y,makeacol_depth(32,r,g,b,alpha));
			}
		}
	}
	wakeSprite = create_bitmap(24,12);
	clear_to_color(wakeSprite,0);
	for (int i=0; i<100; i++) {
		int x = 8; int y = 4;
		int b = 255;
		for (int j=0; j<4; j++) {
			x+= (rand() % 7) - 3;
			y+= (rand() % 5) - 2;
			putpixel(wakeSprite, x,y, makecol(b,b,b));
			b -=40;
		}
	}
}

void HoverCraft::clearSprites()
{
	if (sprites != 0) {
		for (int i=0; i<36; i++) {
			destroy_bitmap(sprites[i]);
		}
		delete [] sprites;
		sprites = 0;
	}
	if (wakeSprite) {
		destroy_bitmap(wakeSprite);
	}
}

HoverCraft::~HoverCraft()
{
	clearSprites();
	setController(0);
}

static const Component CentreHeight = 20;

void HoverCraft::renderTo(const View &v)
{
	int sprite = (int) (azimuth / 10);
	sprite = (sprite + 27) % 36;
	BITMAP * spritebm = sprites[sprite];
	
	Point p = v.translate(location);
	p.x -= ((spritebm->w) / 2);
	p.y -= (spritebm->h);
	p.y += CentreHeight;
	set_alpha_blender();
	draw_trans_sprite(v.bitmap, spritebm, (int) p.x, (int) p.y);
	// debug : render right vector
	/*
	int white = makecol(255,255,255);
	p = v.translate(location);
	Point p2 = v.translate(location + (right * 20));
	line(v.bitmap, (int) p.x, (int) p.y, (int) p2.x, (int) p2.y, white);
	*/
}

static const Component FrictionFactor = 0.04;
static const Component EnginePower = 30.0;
static const Component SpinSpeed = 20.0;
static const Component RotationDamping = 0.1;
static const Component GradientFactor = 100.0;

void HoverCraft::calcVectors()
{
	forwards = Vect3d(cos(toRadians(azimuth)), - sin(toRadians(azimuth)), 0);
	// Manual hack: rotate by 90 deg
	right.y = -forwards.x;
	right.x = forwards.y;
	right.z = 0;
	frontBumper = location + (forwards * BumperDist);
	rearBumper = location - (forwards * BumperDist);
}

void HoverCraft::boundCheck()
{
	if ((location.x <0) || (location.y < 0) ||
	 	(location.x > Map::BlocksWide * Map::BlockSize) ||
	 	(location.y > Map::BlocksWide * Map::BlockSize))
	{
		location = lastCheckpointLocation;
		frozen = true;
		// Long enough to hopefully discourage them from deliberately
		// going out of bounds.
		powerupTime = 6.0;
		if (isPlayer) {
			manager->showBigMessage(language::current->OutOfBounds);
			play_sample(manager->getSample(OHNO_WAV), 192, 128, 1000, 0);
		}
	}
}

static const Component MaxWakeTime = 8.0;

void HoverCraft::createWake(const Vect3d &relPos)
{
	if (manager->getMap()->isWater(Point(location.x, location.y))) {
		Component distance = relPos.magnitude();
		wakeTimer -= distance;
		if (wakeTimer < 0) {
			// create a wake object...
			Wake *wake = new Wake(manager, wakeSprite);
			wake->setLocation(location);
			wake->setVelocity(velocity);
			manager->addObject(wake);
			wakeTimer = MaxWakeTime;
		}
	}
}

void HoverCraft::tick(Component quantum)
{
	if (controller) {
		controller->tick(quantum);
	}
	if (powerupTime > 0) {
		powerupTime -= quantum;
		if (powerupTime < 0) {
			clearPowerups();
		}
	}
	if (frozen) {
		velocity = Vect3d(0,0,0);
		velocityDelta = Vect3d(0,0,0);
		return;
	}
	// Add the velocity from any collisions since last tick...
	velocity = velocity + velocityDelta;
	velocityDelta = Vect3d(0,0,0);
	// Process forces on centre.
	Vect3d normal(manager->getMap()->getNormal(Point(location.x, location.y)));
	// Take the components of the normal in x,y
	normal.z = 0;
	Vect3d force(normal * GradientFactor);
	// Apply some friction...
	force = force + (- velocity * FrictionFactor);
	// Work out the force from the engine
	Component power = 0;
	if (controller) power =  controller->getPower();
	if (wonkyFlag) power = 1;
	power = power * (1.0 + (0.5 * turbo)); // Add in the turboes.
	force = force + (forwards * (power * EnginePower));
	
	// Apply acceleration...
	velocity = velocity + (force * quantum); // (divided by mass)
	// Apply velocity...
	location = location + (velocity * quantum);
	createWake(velocity * quantum);
	// Handle rotation...
	Component spinFactor = (0.2 + power);
	Component rudder = controller->getRudder();
	if (wonkyFlag) {
		rudder = 0.5;
	}
	if (controller) rotation = rotation + (rudder * SpinSpeed * spinFactor);
	rotation = rotation * (1.0 - RotationDamping);
	azimuth = azimuth + (rotation * quantum);
	if (azimuth > 360.0) azimuth -= 360.0;
	if (azimuth < 0.0) azimuth += 360.0;
	
	// Make sure we stay glued to the floor...
	location.z = manager->getMap()->getHeight(Point(location.x, location.y));
	calcVectors();
	// Check we haven't gone awry
	boundCheck();
}

void HoverCraft::setLocation(const Vect3d &where)
{
	location = where; // TODO: anything else...
	if (lastCheckpoint == -1) { // just in case...
		lastCheckpointLocation = where;
	}
	calcVectors();
}

void HoverCraft::setController(Controller *c)
{
	if (controller) {
		delete controller;
	}
	controller = c;
}

bool HoverCraft::checkBump(const Vect3d &centre, const Vect3d &otherVelocity, Component radius, bool front
	, Component multiplier)
{
	// Check bump with bumper (front or rear)
	Vect3d &bumperPos = front ? frontBumper : rearBumper;
	Vect3d relPos = centre - bumperPos;
	Component distSquared = relPos.magnitudeSquared();
	Component touchingSquared = (BumperRadius + radius);
	touchingSquared = touchingSquared * touchingSquared;
	if (distSquared <  touchingSquared ) {
		// work out direction of collision.
		Vect3d collisionDir = relPos;
		collisionDir.normalise();
		// Work out the component of the collision in opposition to the 
		// relative velocity...
		Vect3d relVelocity = otherVelocity - velocity;
		Vect3d relVelocityDir = relVelocity;
		relVelocityDir.normalise();
		Component power = collisionDir.dotProduct(relVelocity);
		// std::cout << "Player: " << playerNum << "Collision power: " << power << std::endl;
		// That doesn't make sense, if the collision is *towards* the rel velocity!
		// It should be in opposition
		if (power > 0) return false;
		// Collision velocity
		Vect3d collisionVelocity = (collisionDir * power ) * multiplier;
		// Apply this to our own velocity
		velocityDelta = velocityDelta + collisionVelocity;
		return true;
	}
	return false;
}

void HoverCraft::handleHoverCraftCollision(HoverCraft *otherCraft)
{
	if (checkBump(otherCraft->frontBumper, otherCraft->velocity, BumperRadius, false, 1))
		return;
	if (checkBump(otherCraft->frontBumper, otherCraft->velocity, BumperRadius, true, 1)) 
		return;
	if (checkBump(otherCraft->rearBumper, otherCraft->velocity, BumperRadius, false, 1))
		return;
	if (checkBump(otherCraft->rearBumper, otherCraft->velocity, BumperRadius, true, 1))
		return;
}

void HoverCraft::checkPoint(CheckPoint *cp)
{
	// Determine whether we've actually crossed it...
	Vect3d relative = location - cp->getLocation();
	// Are we behind the line?
	// Give up.
	if (relative.dotProduct(cp->getForward()) <0) 
		return;
	// Yay, we've crossed the line...
	// Determine if it's the RIGHT checkpoint...
	if (cp->getNumber() != nextCheckpoint)
		return; // Wrong way, dude!
	if (courseComplete) // Ignore checkpoints once we have finished.
		return; 
	// Ok, WE'VE HIT A CHECKPOINT!
	lastCheckpointTime = manager->getTime();
	lastCheckpoint = nextCheckpoint;
	lastCheckpointLocation = location;
	
	if (nextCheckpoint == 0) {
		// NEW LAP!
		currentLap ++;
		if (currentLap > 1) { // Ignore lap zero, as it was an "outlap", 
			// just from the grid to start/finish
			lastLapTime = manager->getTime() - currentLapBegin;
		}
		currentLapBegin = manager->getTime();
		if (currentLap > manager->getNumLaps()) {
			courseComplete = true;
		}
		using namespace std;
		ostringstream oss;
		if (! courseComplete) {
	        	oss << language::current->Lap << " " << currentLap;
		} else {
			oss << language::current->Finished;
		}
		if (isPlayer) {
			manager->showBigMessage(oss.str());
		}
	}
	nextCheckpoint ++;
	nextCheckpoint = nextCheckpoint % manager->getNumCheckpoints();
	manager->sortPlaces();
}

void HoverCraft::handleCollision(GameObject *obj)
{
	// If the obj is a hovercraft, handle specially
	HoverCraft *otherCraft = dynamic_cast<HoverCraft *>(obj);
	if (otherCraft != 0) {
		handleHoverCraftCollision(otherCraft);
	} else {
		// otherwise do a normal collision detection.
		if (controller) {
			controller->notifyCollision(obj);
		}
		
		if (obj->isSolid()) {
			if (checkBump(obj->getLocation(), Vect3d(0,0,0), obj->getRadius(), true, 2))
				return;
			if (checkBump(obj->getLocation(), Vect3d(0,0,0), obj->getRadius(), false, 2))
				return;
		}
		CheckPoint *cp = dynamic_cast<CheckPoint *>(obj);
		if (cp) {
			checkPoint(cp);
		}
	}
}

static void formatTime(std::ostream &os, long ms) 
{
	using namespace std;
	os.fill('0');
	os.width(2);
	os << ms / (60 * 1000);
	os << ':';
	os.width(2);
	os << (ms / ((1000)) % 60);
	os << '.';
	os.width(3);
	os << (ms % 1000);
}

void HoverCraft::renderHud(BITMAP *bm)
{
	using namespace std;
	int white = makecol(255,255,255);
	// Show things like
	// Total time
	ostringstream oss;
	long lapTime = manager->getTime() - currentLapBegin;
        oss << "lap:" << currentLap;
	oss << " current lap time:";
	formatTime(oss,lapTime);
	oss << " last lap time:";
	formatTime(oss,lastLapTime);
	oss << " sector:" << nextCheckpoint <<
		" P" << rank;
	textout(bm, font, oss.str().c_str(), 5,5, white);
	
	// Lap counter
	// Lap time
	// Sector counter
	// Checkpoint time
	// Show what powerups are in effect.
	// At least show turboes.
	for (int i=0; i<turbo; i++) {
		draw_sprite(bm, manager->getSprite(TURBO_TGA), 600, i * 8);
	}
	if (wonkyFlag) 
		draw_sprite(bm, manager->getSprite(CRAZY_TGA), 600, 1);
	if (frozen) 
		draw_sprite(bm, manager->getSprite(ICECUBE_TGA), 600, 1);
}

// powerup effects

static const Component FreezeTime = 8.0;
static const int MaxTurbo = 2;
static const Component TurboTime = 16.0;

void HoverCraft::freeze()
{
	turbo = 0;
	wonkyFlag = false;
	frozen = true;
	powerupTime = FreezeTime;
	if (isPlayer) {
		manager->showBigMessage(language::current->Frozen);
		play_sample(manager->getSample(OHNO_WAV), 192, 128, 1000, 0);
	}
}

void HoverCraft::wonky()
{
	wonkyFlag = true;
	powerupTime = FreezeTime;
	if (isPlayer) 
	{
		manager->showBigMessage(language::current->Wonky);
		play_sample(manager->getSample(OHNO_WAV), 192, 128, 1000, 0);
	}
}

void HoverCraft::addTurbo()
{
	frozen = false;
	wonkyFlag = false;
	if (turbo > 0) {
		powerupTime += TurboTime;
	} else {
		powerupTime = TurboTime;
	}
	turbo ++;
	if (turbo > MaxTurbo) turbo = MaxTurbo;
	if (isPlayer) {
		manager->showBigMessage(language::current->Turbo);
		play_sample(manager->getSample(TING_WAV), 192, 128, 1000, 0);
	}
}

void HoverCraft::clearPowerups()
{
	frozen = false;
	wonkyFlag = false;
	turbo = 0;
}

// Wake
void Wake::renderTo(const View &v)
{
	int alpha = (int) (life * 128);
	if (alpha>128) alpha = 128;
	set_add_blender(0,0,0,alpha);
	Point p = v.translate(location);
	p.x -= ((spritebm->w) / 2);
	p.y -= ((spritebm->h) / 2);
	draw_trans_sprite(v.bitmap, spritebm, (int) p.x, (int) p.y);
	slowTimer = 0;
}

void Wake::tick(Component quantum)
{
	if (slowTimer < 0.01) {
		velocity = (velocity * 0.9);
		slowTimer = 0.5;
	} else {
		slowTimer -= quantum;
	}
	location = location + (velocity * quantum);
	life = life - quantum;
	if (life <0) dead = true;
}
