#ifndef HOVERCRAFT_INCLUDED
	#define HOVERCRAFT_INCLUDED

#include "gameobject.h"

class Controller {
	public:
		virtual void tick(Component quantum) = 0;
		// Returns something between -1 and 1
		virtual Component getRudder() { return 0; }
		// Returns something between 0 and 1
		virtual Component getPower() { return 0; }
		virtual void notifyCollision(GameObject *obj) { }
};

class KeyboardController : public Controller {
	private:
		Component power, rudder;
	public:
		virtual void tick(Component quantum);
		// Returns something between -1 and 1
		virtual Component getRudder();
		// Returns something between 0 and 1
		virtual Component getPower();
};

class HoverCraft: public GameObject {
	private:
		Component azimuth; // degrees
		Vect3d velocity; // units per second
		Component rotation; // degrees per second
		Controller *controller;
		BITMAP **sprites;
		BITMAP *wakeSprite;
		Component wakeTimer;
		void initSprites();
		void clearSprites();
		void modifyColour(int &r, int &g, int &b);
		int playerNum;
		// These hold the positions of the front and rear bumpers,
		// which are spherical objects which are used to process
		// collisions
		Vect3d frontBumper;
		Vect3d rearBumper;
		Vect3d forwards;
		Vect3d right;
		Vect3d velocityDelta;
		Vect3d lastCheckpointLocation; // Just in case they fall off and we need to put the back.
		// Data for the Timing / checkpoints system
		int nextCheckpoint;
		int lastCheckpoint;
		long lastCheckpointTime;
		long lastLapTime; // Total
		long currentLapBegin; // ms time that the current lap began.
		int currentLap; // 0 = not yet crossed start / finish line 1st time
		int rank;
		bool courseComplete;
		
		// powerups
		Component powerupTime;
		int turbo; // number of turbos
		bool wonkyFlag;
		bool frozen;
		bool isPlayer;
		
		void calcVectors();
		bool checkBump(const Vect3d &centre, const Vect3d &otherVelocity, 
			Component radius, bool front, Component multiplier);
		void handleHoverCraftCollision(HoverCraft *otherCraft);
		void checkPoint(CheckPoint *cp);
		void clearPowerups();
		void boundCheck();
		void createWake(const Vect3d &relPos);
		
	public:
		HoverCraft(Manager *m, int p);
		virtual ~HoverCraft();
		virtual void renderTo(const View &v);
		virtual void tick(Component quantum);
		virtual void setLocation(const Vect3d &where);
		void setController(Controller *c);
		void handleCollision(GameObject *obj);
		void renderHud(BITMAP *bm);
		Component getAzimuth() { return azimuth; }
		void setAzimuth(Component a) { azimuth = a; calcVectors(); }
		Vect3d getVelocity() { return velocity; }
		void setVelocity(const Vect3d &v) { velocity = v; }
		int getCurrentLap() { return currentLap; }
		int getLastCheckpoint() { return lastCheckpoint; }
		long getLastCheckpointTime() { return lastCheckpointTime; }
		void setRank(int r) { rank = r; }
		int getRank() { return rank; }
		// powerup effects
		void freeze();
		void wonky();
		void addTurbo();
		void setIsPlayer(bool b) { isPlayer = b; }
};

class Wake : public GameObject {
	private:
		BITMAP *spritebm;
		Vect3d velocity;
		Component life;
		Component slowTimer;
	public:
		Wake(Manager *m, BITMAP *bm) : GameObject(m) { 
			spritebm = bm; life = 1.2; radius = 16; 
			collidable = false;
		}
		virtual ~Wake() { }
		virtual void renderTo(const View &v);
		virtual void tick(Component quantum);
		virtual void setLocation(const Vect3d &where) { location = where; }
		void setVelocity(const Vect3d &v) { velocity = v; }
};


#endif // HOVERCRAFT_INCLUDED

